#!/usr/bin/python3 -v

from datetime import datetime
from datetime import date
from datetime import timezone
from datetime import timedelta
import json
import os
import sys
import mysql.connector
import smtplib
import time

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

STOR="/var/wikiwatch.json"

SMTPSERV="localhost"
SMTPFROM="wikiwatch@fidoman.ru"

SUSPEND=10

DB_CONN = lambda: mysql.connector.connect(user='wikiwatch', password='Qwer0987', database='kbmw26')

def cmp_dicts(d1, d2):
  k1 = set(d1.keys())
  k2 = set(d2.keys())
  added = k2-k1
  removed = k1-k2
  tocompare = k1.intersection(k2)
  return added, removed, tocompare

default_messages = {
'new_category':         "new category: %(category)s pages: %(page)s",
'removed_category':     "removed category: %(category)s",
'page_added':           "page %(page)s added to category %(category)s",
'page_deleted':         "page %(page)s deleted from category %(category)s",
'page_new_in_category': "new page %(page)s in category %(category)s",
'page_new':             "new page %(page)s without any category",
'page_edit':            "page %(page)s in category %(category)s was edited",
'page_edit_tmstmp':     "page %(page)s in category %(category)s was edited on %(tmstmp)s",
'page_removed':         "page %(page)s removed"
}

default_NSs = {
    '0': '',
    '1': 'Talk',
    '2': 'User',
    '3': 'User_talk',
    '4': 'Project',
    '5': 'Project_talk',
    '6': 'File',
    '7': 'File_talk',
    '8': 'MediaWiki',
    '9': 'MediaWiki_talk',
    '10': 'Template',
    '11': 'Template_talk',
    '12': 'Help',
    '13': 'Help_talk',
    '14': 'Category',
    '15': 'Category_talk',
}




def anchorize(p, baseurl):
  pa = p.copy()
  if baseurl:
    if "page" in pa:
      pa["page"] = '<a href="'+baseurl + pa["page"] + '">' + pa["page"] + "</a>"
    if "category" in pa:
      pa["category"] = '<a href="'+baseurl + 'Category:' + pa["category"] + '">' + pa["category"] + "</a>"
  return pa

class Notificator:
  """Keeps data about emails and aggregates notifications"""
  def __init__(self, addresses, templates, baseurl):
    self.messages = {} # email -> [data lines]
    self.templates = templates
    self.cat_emails=addresses
    self.baseurl = baseurl
    #print("Notify emails:", self.cat_emails)
    
  def add(self, category, message, params):
    emails=self.cat_emails.get(category)
    if not emails: # None or []
      print("Notification for category '%s' will not be sent [%s]"%(category, message))
      return
    tpl = self.templates.get(message)
    if not tpl:
      print("Notification will not be sent: no template for message '%s'"%message)
      return
    #msg_text=tpl%params
      
    for email in emails:
      #print("Notify %s: %s"%(email, message))
      self.messages.setdefault(email, []).append((tpl, params))

  def send_all(self, subject="wiki watch notification"):
    for dst in self.messages.keys():
#      try:
        print("sending to:", dst)
        s=smtplib.SMTP("localhost")
      
        msg = MIMEMultipart('alternative')
        msg['Subject'] = subject 
        msg['From'] = "<"+SMTPFROM+">"
        msg['To'] = "<"+dst+">"
      
        # self.messages[dst] is list of tuples to write in body
        msg_plaintext = map(lambda x: x[0]%x[1], self.messages[dst])
        # for HTML: add <a> to param values
        msg_html = map(lambda x: x[0]%(anchorize(x[1], self.baseurl)), self.messages[dst])
      
        body = '\n'.join(msg_plaintext)
        part = MIMEText(body, 'plain')
        msg.attach(part)
        
        body = '<br>\n'.join(msg_html)
        part = MIMEText(body, 'html')
        msg.attach(part)
      
        #print(msg.as_string())
      
        res=s.sendmail(SMTPFROM, dst, msg.as_string())
        s.quit()
        print("SMTP result:", res)
#      except Exception as e:
#        print("send to", dst, "failed:", e)
        



#START:
#if exist .new file but not exist current: move # terminated after successful save and deletion 
#if exist current: discard .new if exist # failed before deletion OR completion of write

#END:
#  save data to .new
#  move current to .old with overwrite
#  rename new to current

class ActivityStor:
  """our memory about pages and categories"""
  def __init__(self, storfilename):
    self.fname = storfilename
    self.lockf=None
    
  def __enter__(self):    
    self.lockf=os.open(self.fname+".lock", os.O_CREAT | os.O_EXCL)
    print("lock acquired for", self.fname)

    if not os.path.exists(self.fname):
      if os.path.exists(self.fname+".new"):
        os.rename(self.fname+".new", self.fname)
      else:
        print("initializing data file")
        x=open(self.fname, "w")
        json.dump(None, x)
        x.close()
      
    dataf = open(self.fname)
    self.data = json.load(dataf)
    dataf.close()
    #print("READ:", repr(self.data))
    return self

  def commit(self):
    dataf=open(self.fname+".new", "w")
    json.dump(self.data, dataf)
    dataf.close()
    print("data flushed")
    os.unlink(self.fname)
    os.rename(self.fname+".new", self.fname)

  def __exit__(self, et, ev, etb):
    if et:
      print("exiting", str(et), ev, etb)
    if self.lockf is not None:
        #self.commit()
        print("releasing lock for", self.fname)
        os.close(self.lockf)
        os.unlink(self.fname+".lock")


def pagename(NSs, pagens, title):
      nsname = NSs.get(pagens)
      if nsname is None:
        return "NS#%s:%s"%(pagens, title) # unknown namespace number
      if nsname=='': # MAIN
        return "%s"%(title)
      return "%s:%s"%(nsname, title)



def diff():
  ntf = Notificator(load_addresses(), load_messages(), load_urlbase())
  NSs = load_namespaces()

  with ActivityStor(STOR) as x:
    if x.data is None:
      x.data=[{}, {}] # pages, categories

    db_catmembers={}
    db_pages={}
    db_pagecats={}



    a = DB_CONN()
    dbdata=a.cmd_query('select page_id, page_namespace, page_title, page_touched, cl_to from page, categorylinks where page_id=cl_from')
    # Для использования recent_changes надо поменять логику - отслеживать последнюю обработанную записть в таблице, иначе размер ответа разрастается
#    dbdata=a.cmd_query('select page_id, page_namespace, page_title, rc_timestamp, cl_to from page, categorylinks, recentchanges where page_id=cl_from and rc_namespace=page_namespace and rc_title=page_title')
    #print(dbdata)
    while True:
      d=a.get_row()
      #print(d[0],d[1])
      if d[0] is None:
        break
      pgid=d[0][0].decode("utf-8")
      pgns=d[0][1].decode("utf-8")
      pgtitle=d[0][2].decode("utf-8")
      pgtmstmp=datetime.strptime(d[0][3].decode("ascii")+"+0000", "%Y%m%d%H%M%S%z").timestamp() # assume UTC in database
      pgcat=d[0][4].decode("utf-8")
        
      db_pages[pgid] = [pgns, pgtitle, pgtmstmp] # can be moved to separate query; use list as json make tuple lists and brokes comparation
      db_catmembers.setdefault(pgcat, []).append(pgid)
      db_pagecats.setdefault(pgid, []).append(pgcat)

    # add virtual category membership for category pages in themselves
    dbdata=a.cmd_query('select page_id, page_namespace, page_title, page_touched, cat_title from page, category where page_namespace=14 and page_title=cat_title')
    #print(dbdata)
    while True:
      d=a.get_row()
      #print(d[0],d[1])
      if d[0] is None:
        break
      pgid=d[0][0].decode("utf-8")
      pgns=d[0][1].decode("utf-8")
      pgtitle=d[0][2].decode("utf-8")
      pgtmstmp=datetime.strptime(d[0][3].decode("ascii")+"+0000", "%Y%m%d%H%M%S%z").timestamp() # assume UTC in database
      pgcat=d[0][4].decode("utf-8")
      
      print("virtual membership for", pgns, pgtitle, "in", pgcat)

      db_pages[pgid] = [pgns, pgtitle, pgtmstmp] # can be moved to separate query; use list as json make tuple lists and brokes comparation
      db_catmembers.setdefault(pgcat, []).append(pgid)
      db_pagecats.setdefault(pgid, []).append(pgcat)


    #print("Check categories")
    cat_added, cat_removed, cat_check = cmp_dicts(x.data[1], db_catmembers)

    #... notify about new and removed categories, new - with page list
    if cat_added:
      print("new categories:", cat_added)
      #messages.setdefault(cat_meta_email, []).append("new categories: %s"%cat_added)
      for c1 in cat_added:
        pagelist = []
        for cp1 in db_catmembers[c1]:
          pagelist.append(db_pages[cp1][1]) #!!! add namespace
        ntf.add('', "new_category", {"category": c1, "pages": ", ".join(pagelist)})
        x.data[1][c1]=db_catmembers[c1] # copy catmembers to our stor
    if cat_removed:
      print("removed categories:", cat_removed)
      for c1 in cat_removed:
        ntf.add('', "removed_category", {"category": c1, "pages": ""})
        x.data[1].pop(c1)

    #... check changed categories
    for c1 in cat_check:
      #print("check membership change for", c1)
      #print("new", db_catmembers[c1], "old", x.data[1][c1])
      added=set(db_catmembers[c1])-set(x.data[1][c1])
      removed=set(x.data[1][c1])-set(db_catmembers[c1])
    
      if added or removed:
        print("changes in",c1,":", added, removed)
      
      for cp1 in added:
        ntf.add(c1, "page_added", {"category": c1, "page": pagename(NSs, *db_pages[cp1][0:2])})
        x.data[1][c1].append(cp1)

      for cp1 in removed:
        ntf.add(c1, "page_deleted", {"category": c1, "page": pagename(NSs, *db_pages[cp1][0:2])})
        x.data[1][c1].remove(cp1)


    #print("Checking pages")
    #print(x.data[0], db_pages)

    # should replace page_touch compare with recent_change list
    # keep latest viewed recent_changes mark in data file

    pg_added, pg_removed, pg_check = cmp_dicts(x.data[0], db_pages)
    if pg_added:
      for p1 in pg_added:
        print("added", p1, db_pagecats[p1])
        for c1 in db_pagecats[p1]:
          ntf.add(c1, "page_new_in_category", {"page": pagename(NSs, *db_pages[p1][0:2]), "category": c1})
        if not db_pagecats[p1]:
          ntf.add('', "page_new", {"page": pagename(NSs, *db_pages[p1][0:2]), "category": ""})
        x.data[0][p1] = db_pages[p1] # +[] if additional data is needed

    if pg_removed:
      for p1 in pg_removed:
        print("removed", p1) # search categories in x.data[1]
        x.data[0].pop(p1)

    for p1 in pg_check:
      # intersection - compare time
      #print("checking modify time for page", p1)
      oldtmstmp=x.data[0][p1][2]
      newtmstmp=db_pages[p1][2]
      curtmstmp=datetime.now(timezone.utc).timestamp()
      # notify if: 1) new>old and cur>new for more than 45 min; ?2) cur>old for more that 3 hrs
      #print(oldtmstmp, newtmstmp, curtmstmp)
      if newtmstmp>oldtmstmp:
        print("page", p1, "changed %ds ago"%(curtmstmp-newtmstmp))
        if curtmstmp-newtmstmp < SUSPEND:
          print("suspend")
        else:
          print("last edit was long ago!")
          for c1 in db_pagecats[p1]:
            print("the page belongs to category", c1)
            ntf.add(c1, "page_edit", {"category": c1, "page": pagename(NSs, *db_pages[p1][0:2])})
            x.data[0][p1][2] = newtmstmp
      else:
        pass # no new changes

    ntf.send_all()
    #print("updating stor")
    x.commit()



def summary(days):
  """find changes from midnight `days` days ago up to current time"""
  NSs = load_namespaces()

  curdt=datetime.now(timezone.utc)
  #print(curdt)
  tm=curdt.time()
  d=timedelta(days, hours=tm.hour, minutes=tm.minute, seconds=tm.second, microseconds=tm.microsecond)
  startdt=curdt-d
  starttmstmp=startdt.timestamp()
  endtmstmp=curdt.timestamp()
  #print(starttmstmp, endtmstmp)
  
  ntf = Notificator(load_addresses(), load_messages(), load_urlbase())

  with ActivityStor(STOR) as x:
    for p1 in x.data[0]:
      #print(p1, x.data[0][p1])
      if x.data[0][p1][2]>starttmstmp:
        ascts=time.ctime(x.data[0][p1][2])
        print(p1, "was edited", ascts)
        for c1 in x.data[1]:
          if p1 in x.data[1][c1]:
            pgname = pagename(NSs, *x.data[0][p1][0:2])
            ntf.add(c1, "page_edit_tmstmp", {"page": pgname, "category": c1, "tmstmp": ascts})

    ntf.send_all("wiki edits in last %d days"%days)

def run(slp=500):
  while True:
    diff()
    time.sleep(slp)


def usage():
  print("""Mediawiki change watcher
  watcher.py clearlock          -- remove stale lock file (run it on system startup)
  watcher.py summary N          -- notify about changes in last N days (midnight to midnight)
  watcher.py diff               -- notify about changes since last diff
  watcher.py run [N]            -- run diff each N seconds
  watcher.py emails             -- show list of e-mails for categories
  watcher.py messages           -- show list of localized messages
  watcher.py namespaces         -- show list of custom namespaces
  watcher.py urlbase            -- show base URL for mediawiki pages
""")





def load_MW_page_dict(pagename):
  """ read dict from Mediawiki namespace page """
  a = DB_CONN()
  a.cmd_query("select old_text from text, revision, page where page_namespace=8 and page_title='%s' and page_latest=rev_id and rev_text_id=old_id"%pagename) # vulnerable!!! but no positional parameters in this function is supported
  pgtext = None
  while True:
    d, stat = a.get_row()
    #print(d, stat)
    if not d:
      break
    pgtext = d[0].decode("utf-8")
  if pgtext is None:
    print("cannot find page Category_notify in wiki database")
    exit()
    
  addrs = {}
  
  cat = None
  for l1 in pgtext.strip().split("\n"):
    if l1.startswith(';'): # new 'definition'
      r = list(map(str.strip, l1[1:].split(':', 1)))
      if len(r)!=2:
        print( "bad e-mail record", l1)
        continue
      cat, email = r
      #print(cat, email)
      if cat=='*':
        cat=''
      cat = cat.replace(' ', '_')
      addrs[cat] = [email]
    elif l1.startswith(':'): # additional address
      email = l1[1:].strip()
      if cat is None:
        print( "orphan e-mail", l1)
        continue
      addrs[cat].append(email)

  return addrs

def load_addresses():
  return load_MW_page_dict('Category_notify')

def load_messages():
  msgs = default_messages.copy()
  for k, v in load_MW_page_dict('Category_notify_templates').items():
    msgs[k] = "\n".join(v)

  return msgs


def show_emails():
  for k, v in load_addresses().items():
    print("cat", repr(k), "email", v)

def show_messages():
  for k, v in load_messages().items():
    print("message  :", repr(k))
    print("localized:", v)
    print()


def load_namespaces():
  NSs = default_NSs.copy()
  with open("vars.json") as vfile:
    v=json.load(vfile)
  NSs.update(v.get("wgExtraNamespaces", {}))
  return NSs
  
def show_namespaces():
  for k, v in load_namespaces().items():
    print(repr(k), repr(v))

def load_urlbase():
  with open("vars.json") as vfile:
    v=json.load(vfile)
  server = v.get('wgServer')
  path = v.get('wgScriptPath')
  return server+path+'/index.php?title='
  


if len(sys.argv)>1 and sys.argv[1]=="clearlock":
  try:
    os.unlink(STOR+".lock")
    print("lock file removed")
  except:
    pass
  exit()

if len(sys.argv)>2 and sys.argv[1]=="summary":
  try:
    days=int(sys.argv[2])
  except:
    usage()
    exit()
  summary(days)
  exit()

if len(sys.argv)>1 and sys.argv[1]=="diff":
  diff()
  exit()

if len(sys.argv)==2 and sys.argv[1]=="run":
  run()
  exit()

if len(sys.argv)==3 and sys.argv[1]=="run":
  run(int(sys.argv[2]))
  exit()

if len(sys.argv)>1 and sys.argv[1]=="emails":
  show_emails()
  exit()

if len(sys.argv)>1 and sys.argv[1]=="namespaces":
  show_namespaces()
  exit()

if len(sys.argv)>1 and sys.argv[1]=="messages":
  show_messages()
  exit()

if len(sys.argv)>1 and sys.argv[1]=="urlbase":
  print(load_urlbase())
  exit()

  
usage()
exit()
